from .paginator import Paginator


def test_paginator():

    items = range(1, 491)
    p = Paginator(items, page=1, per_page=20)

    assert p.page == 1
    assert not p.has_prev
    assert p.has_next
    assert p.total_items == 490
    assert p.total_pages == 25
    assert p.next_page_number == 2
    assert list(p.pages) == [1, 2, 3, 4, None, 24, 25]
    assert p.pages_range == (0, 19)

    p.page = 10
    assert p.page == 10
    assert p.has_prev
    assert list(p.pages) == [1, 2, None, 7, 8, 9, 10, 11, 12, 13, None, 24, 25]
    assert list(p) == list(range(181, 201))

    assert Paginator(range(5))



from sqlalchemy.orm import Query
from sqlalchemy_tools.pagination import Paginator


class BaseQuery(Query):

    def get_or_error(self, uid, error):
        """Like :meth:`get` but raises an error if not found instead of
        returning `None`.
        """
        rv = self.get(uid)
        if rv is None:
            if isinstance(error, Exception):
                raise error
            return error()
        return rv

    def first_or_error(self, error):
        """Like :meth:`first` but raises an error if not found instead of
        returning `None`.
        """
        rv = self.first()
        if rv is None:
            if isinstance(error, Exception):
                raise error
            return error()
        return rv

    def paginate(self, **kwargs):
        """Paginate this results.
        Returns an :class:`Paginator` object.
        """
        return Paginator(self, **kwargs)



import argparse
from functools import wraps
import logging
import os
import sys
from flask import current_app
from manager import Manager
from alembic import __version__ as __alembic_version__
from alembic.config import Config as AlembicConfig
from alembic import command
from alembic.util import CommandError

alembic_version = tuple([int(v) for v in __alembic_version__.split('.')[0:3]])
log = logging.getLogger(__name__)


class _MigrateConfig(object):
    def __init__(self, migrate, db, **kwargs):
        self.migrate: Migrate = migrate
        self.db = db
        self.directory = migrate.directory
        self.configure_args = kwargs

    @property
    def metadata(self):
        """
        Backwards compatibility, in old releases app.extensions['migrate']
        was set to db, and env.py accessed app.extensions['migrate'].metadata
        """
        return self.db.metadata


class Config(AlembicConfig):
    def get_template_directory(self):
        package_dir = os.path.abspath(os.path.dirname(__file__))
        return os.path.join(package_dir, 'templates')


class Migrate(object):
    def __init__(self, db, app=None, directory='migrations', **kwargs):
        self.configure_callbacks = []
        self.db = db
        self.directory = str(directory)
        self.alembic_ctx_kwargs = kwargs
        self.config = _MigrateConfig(
            self, self.db, **self.alembic_ctx_kwargs)
        if app is not None and db is not None:
            self.init_app(app, db, directory)

    def init_app(self, app, db=None, directory=None, **kwargs):
        self.db = db or self.db
        self.directory = str(directory or self.directory)
        self.alembic_ctx_kwargs.update(kwargs)
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        app.extensions['migrate'] = self.config

    def configure(self, f):
        self.configure_callbacks.append(f)
        return f

    def call_configure_callbacks(self, config):
        for f in self.configure_callbacks:
            config = f(config)
        return config

    def get_config(self, directory=None, x_arg=None, opts=None):
        if directory is None:
            directory = self.directory
        directory = str(directory)
        config = Config(os.path.join(directory, 'alembic.ini'))
        config.set_main_option('script_location', directory)
        if config.cmd_opts is None:
            config.cmd_opts = argparse.Namespace()
        for opt in opts or []:
            setattr(config.cmd_opts, opt, True)
        if not hasattr(config.cmd_opts, 'x'):
            if x_arg is not None:
                setattr(config.cmd_opts, 'x', [])
                if isinstance(x_arg, list) or isinstance(x_arg, tuple):
                    for x in x_arg:
                        config.cmd_opts.x.append(x)
                else:
                    config.cmd_opts.x.append(x_arg)
            else:
                setattr(config.cmd_opts, 'x', None)
        return self.call_configure_callbacks(config)


def catch_errors(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            f(*args, **kwargs)
        except (CommandError, RuntimeError) as exc:
            log.error('Error: ' + str(exc))
            sys.exit(1)
    return wrapped


def get_next_rev_id(directory):
    count = 0
    for filename in os.listdir(directory):
        if not filename.endswith('.py'):
            continue
        count += 1
    return str(count + 1).zfill(4)


class MigrateManager(Manager):
    @property
    def migrate_config(self):
        migrate_config = getattr(self, '_migrate_config', None)
        if not migrate_config:
            raise ValueError("'migrate_config' not set")
        return self._migrate_config

    @migrate_config.setter
    def migrate_config(self, value: _MigrateConfig):
        if isinstance(value, _MigrateConfig):
            self._migrate_config = value
            return self.migrate_config

    def set_migrate(self, migrate: Migrate):
        self.migrate_config = migrate.config


migrate_manager = MigrateManager()


@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.arg('template', flag='template', default='default',
                     help=("Multiple databases migraton (default is False)"))
@migrate_manager.command
@catch_errors
def init(directory=None, template=None):
    """Creates a new migration repository"""
    if directory is None:
        directory = migrate_manager.migrate_config.directory
    config = Config()
    config.set_main_option('script_location', directory)
    config.config_file_name = os.path.join(directory, 'alembic.ini')
    config = migrate_manager.migrate_config.migrate.call_configure_callbacks(config)
    if not os.path.exists(os.path.join(config.get_template_directory(), template)):
        raise ValueError("Template does not exist")
    command.init(config, directory, template)


@migrate_manager.arg('rev_id', flag='rev-id', default=None,
                     help=('Specify a hardcoded revision id instead of generating the next one in sequence'))
@migrate_manager.arg('version_path', flag='version-path', default=None,
                     help=('Specify specific path from config for version file'))
@migrate_manager.arg('branch_label', flag='branch-label', default=None,
                     help=('Specify a branch label to apply to the new revision'))
@migrate_manager.arg('splice', flag='splice', type=bool,
                     default=False,
                     help=('Allow a non-head revision as the "head" to splice onto'))
@migrate_manager.arg('head', flag='head', default='head',
                     help=('Specify head revision or <branchname>@head to base new revision on'))
@migrate_manager.arg('sql', flag='sql', type=bool, default=False,
                     help=("Don't emit SQL to database - dump to standard output instead"))
@migrate_manager.arg('autogenerate', flag='autogenerate',
                     type=bool, default=False,
                     help=('Populate revision script with candidate '
                           'migration operations, based on comparison of '
                           'database to model'))
@migrate_manager.arg('message', flag='message', shortcut='m', default=None,
                     help='Revision message')
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.command
@catch_errors
def revision(directory=None, message=None, autogenerate=False, sql=False,
             head='head', splice=False, branch_label=None, version_path=None,
             rev_id=None):
    """Create a new revision file."""
    config = migrate_manager.migrate_config.migrate.get_config(directory)
    if directory is None:
        directory = migrate_manager.migrate_config.directory
    if rev_id is None:
        rev_id = get_next_rev_id(os.path.join(directory, 'versions'))
    command.revision(config, message, autogenerate=autogenerate, sql=sql,
                     head=head, splice=splice, branch_label=branch_label,
                     version_path=version_path, rev_id=rev_id)


@migrate_manager.arg('rev_id', flag='rev-id', default=None,
                     help=('Specify a hardcoded revision id instead of generating the next one in sequence'))
@migrate_manager.arg('version_path', flag='version-path', default=None,
                     help=('Specify specific path from config for version file'))
@migrate_manager.arg('branch_label', flag='branch-label', default=None,
                     help=('Specify a branch label to apply to the new revision'))
@migrate_manager.arg('splice', flag='splice', type=bool,
                     default=False,
                     help=('Allow a non-head revision as the "head" to splice onto'))
@migrate_manager.arg('head', flag='head', default='head',
                     help=('Specify head revision or <branchname>@head to base new revision on'))
@migrate_manager.arg('sql', flag='sql', type=bool, default=False,
                     help=("Don't emit SQL to database - dump to standard output instead"))
@migrate_manager.arg('message', flag='message', shortcut='m', default=None,
                     help='Revision message')
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.arg('x_arg', flag='x-arg', shortcut='x', default=None,
                     action='append', help=("Additional arguments consumed "
                                            "by custom env.py scripts"))
@migrate_manager.command
@catch_errors
def migrate(directory=None, message=None, sql=False, head='head', splice=False,
            branch_label=None, version_path=None, rev_id=None, x_arg=None):
    """Alias for 'revision --autogenerate'"""
    config = migrate_manager.migrate_config.migrate.get_config(
        directory, opts=['autogenerate'], x_arg=x_arg)
    if directory is None:
        directory = migrate_manager.migrate_config.directory
    if rev_id is None:
        rev_id = get_next_rev_id(os.path.join(directory, 'versions'))
    command.revision(config, message, autogenerate=True, sql=sql,
                     head=head, splice=splice, branch_label=branch_label,
                     version_path=version_path, rev_id=rev_id)


@migrate_manager.arg('revision', nargs='?', default='head',
                     help="revision identifier")
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.command
@catch_errors
def edit(directory=None, revision='current'):
    """Edit current revision."""
    if alembic_version >= (0, 8, 0):
        config = migrate_manager.migrate_config.migrate.get_config(
            directory)
        command.edit(config, revision)
    else:
        raise RuntimeError('Alembic 0.8.0 or greater is required')


@migrate_manager.arg('rev_id', flag='rev-id', default=None,
                     help=('Specify a hardcoded revision id instead of generating one'))
@migrate_manager.arg('branch_label', flag='branch-label', default=None,
                     help=('Specify a branch label to apply to the new revision'))
@migrate_manager.arg('message', flag='message', shortcut='m', default=None,
                     help='Revision message')
@migrate_manager.arg('revisions', nargs='+',
                     help='one or more revisions, or "heads" for all heads')
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.command
@catch_errors
def merge(directory=None, revisions='', message=None, branch_label=None,
          rev_id=None):
    """Merge two revisions together.  Creates a new migration file"""
    config = migrate_manager.migrate_config.migrate.get_config(directory)
    command.merge(config, revisions, message=message,
                  branch_label=branch_label, rev_id=rev_id)


@migrate_manager.arg('tag', flag='tag', default=None,
                     help=("Arbitrary 'tag' name - can be used by custom env.py scripts"))
@migrate_manager.arg('sql', flag='sql', type=bool, default=False,
                     help=("Don't emit SQL to database - dump to standard output instead"))
@migrate_manager.arg('revision', nargs='?', default='head',
                     help="revision identifier")
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.arg('x_arg', flag='x-arg', shortcut='x', default=None,
                     action='append', help=("Additional arguments consumed "
                                            "by custom env.py scripts"))
@migrate_manager.command
@catch_errors
def upgrade(directory=None, revision='head', sql=False, tag=None, x_arg=None):
    """Upgrade to a later version"""
    config = migrate_manager.migrate_config.migrate.get_config(directory,
                                                               x_arg=x_arg)
    command.upgrade(config, revision, sql=sql, tag=tag)


@migrate_manager.arg('tag', flag='tag', default=None,
                     help=("Arbitrary 'tag' name - can be used by custom env.py scripts"))
@migrate_manager.arg('sql', flag='sql', type=bool, default=False,
                     help=("Don't emit SQL to database - dump to standard output instead"))
@migrate_manager.arg('revision', nargs='?', default="-1",
                     help="revision identifier")
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.arg('x_arg', flag='x-arg', shortcut='x', default=None,
                     action='append', help=("Additional arguments consumed "
                                            "by custom env.py scripts"))
@migrate_manager.command
@catch_errors
def downgrade(directory=None, revision='-1', sql=False, tag=None, x_arg=None):
    """Revert to a previous version"""
    config = migrate_manager.migrate_config.migrate.get_config(directory,
                                                               x_arg=x_arg)
    if sql and revision == '-1':
        revision = 'head:-1'
    command.downgrade(config, revision, sql=sql, tag=tag)


@migrate_manager.arg('revision', nargs='?', default="head",
                     help="revision identifier")
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.command
@catch_errors
def show(directory=None, revision='head'):
    """Show the revision denoted by the given symbol."""
    config = migrate_manager.migrate_config.migrate.get_config(directory)
    command.show(config, revision)


@migrate_manager.arg('indicate_current', flag='indicate-current', shortcut='i',
                     type=bool, default=False,
                     help=('Indicate current version (Alembic 0.9.9 or greater is required)'))
@migrate_manager.arg('verbose', flag='verbose', shortcut='v', type=bool,
                     default=False, help='Use more verbose output')
@migrate_manager.arg('rev_range', flag='rev-range', shortcut='r', default=None,
                     help=('Specify a revision range; format is [start]:[end]'))
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.command
@catch_errors
def history(directory=None, rev_range=None, verbose=False,
            indicate_current=False):
    """List changeset scripts in chronological order."""
    config = migrate_manager.migrate_config.migrate.get_config(directory)
    if alembic_version >= (0, 9, 9):
        command.history(config, rev_range, verbose=verbose,
                        indicate_current=indicate_current)
    else:
        command.history(config, rev_range, verbose=verbose)


@migrate_manager.arg('resolve_dependencies', flag='resolve-dependencies', type=bool,
                     default=False, help='Treat dependency versions as down revisions')
@migrate_manager.arg('verbose', flag='verbose', shortcut='v', type=bool,
                     default=False, help='Use more verbose output')
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.command
@catch_errors
def heads(directory=None, verbose=False, resolve_dependencies=False):
    """Show current available heads in the script directory"""
    config = migrate_manager.migrate_config.migrate.get_config(directory)
    command.heads(config, verbose=verbose,
                  resolve_dependencies=resolve_dependencies)


@migrate_manager.arg('verbose', flag='verbose', shortcut='v', type=bool,
                     default=False, help='Use more verbose output')
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.command
@catch_errors
def branches(directory=None, verbose=False):
    """Show current branch points"""
    config = migrate_manager.migrate_config.migrate.get_config(directory)
    command.branches(config, verbose=verbose)


@migrate_manager.arg('head_only', flag='head-only', type=bool,
                     default=False,
                     help='Deprecated. Use --verbose for additional output')
@migrate_manager.arg('verbose', flag='verbose', shortcut='v', type=bool,
                     default=False, help='Use more verbose output')
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.command
@catch_errors
def current(directory=None, verbose=False, head_only=False):
    """Display the current revision for each database."""
    config = migrate_manager.migrate_config.migrate.get_config(directory)
    command.current(config, verbose=verbose, head_only=head_only)


@migrate_manager.arg('tag', flag='tag', default=None,
                     help=("Arbitrary 'tag' name - can be used by custom env.py scripts"))
@migrate_manager.arg('sql', flag='sql', type=bool, default=False,
                     help=("Don't emit SQL to database - dump to standard output instead"))
@migrate_manager.arg('revision', default=None, help="revision identifier")
@migrate_manager.arg('directory', flag='directory', shortcut='d', default=None,
                     help=("Migration script directory (default is 'migrations')"))
@migrate_manager.command
@catch_errors
def stamp(directory=None, revision='head', sql=False, tag=None):
    """'stamp' the revision table with the given revision; don't run any
    migrations"""
    config = migrate_manager.migrate_config.migrate.get_config(directory)
    command.stamp(config, revision, sql=sql, tag=tag)



import threading
from typing import Any, Dict, List

import arrow
import pandas as pd
import sqlalchemy
import sqlalchemy_utils as sa_utils
from arrow import utcnow
from sqlalchemy import *
from sqlalchemy.engine.url import make_url
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Query, make_transient, scoped_session, sessionmaker
from sqlalchemy.schema import MetaData

from .base import BaseModel, BaseQuery


DEFAULT_PER_PAGE = 10


def _create_scoped_session(db, query_cls):
    session = sessionmaker(autoflush=True, autocommit=False,
                           bind=db.engine, query_cls=query_cls)
    return scoped_session(session)


def _tablemaker(db):
    def make_sa_table(*args, **kwargs):
        if len(args) > 1 and isinstance(args[1], db.Column):
            args = (args[0], db.metadata) + args[1:]
        kwargs.setdefault('bind_key', None)
        info = kwargs.pop('info', None) or {}
        info.setdefault('bind_key', None)
        kwargs['info'] = info
        return sqlalchemy.Table(*args, **kwargs)

    return make_sa_table


def _include_sqlalchemy(db):
    for module in sqlalchemy, sqlalchemy.orm:
        for key in module.__all__:
            if not hasattr(db, key):
                setattr(db, key, getattr(module, key))
    db.Table = _tablemaker(db)
    db.event = sqlalchemy.event
    db.utils = sa_utils
    db.arrow = arrow
    db.utcnow = utcnow
    db.SADateTime = db.DateTime
    db.DateTime = sa_utils.ArrowType
    db.JSONType = sa_utils.JSONType
    db.EmailType = sa_utils.EmailType


class EngineConnector:
    def __init__(self, sa_obj):
        self._sa_obj = sa_obj
        self._engine = None
        self._connected_for = None
        self._lock = threading.Lock()

    def get_engine(self):
        with self._lock:
            uri = self._sa_obj.uri
            info = self._sa_obj.info
            options = self._sa_obj.options
            echo = options.get('echo')
            if (uri, echo) == self._connected_for:
                return self._engine
            self._engine = engine = sqlalchemy.create_engine(info, **options)
            self._connected_for = (uri, echo)
            return engine


class Database(object):
    """This class is used to instantiate a SQLAlchemy connection to
    a database.
        db = Database(_uri_to_database_)
    The class also provides access to all the SQLAlchemy
    functions from the :mod:`sqlalchemy` and :mod:`sqlalchemy.orm` modules.
    So you can declare models like this::
        class User(db.Model):
            login = db.Column(db.String(80), unique=True)
            passw_hash = db.Column(db.String(80))
    In a web application you need to call `db.session.remove()`
    after each response, and `db.session.rollback()` if an error occurs.
    If your application object has a `after_request` and `on_exception
    decorators, just pass that object at creation::
        app = Flask(__name__)
        db = Database('sqlite://', app=app)
    or later::
        db = Database()
        app = Flask(__name__)
        db.init_app(app)
    .. admonition:: Check types carefully
       Don't perform type or `isinstance` checks against `db.Table`, which
       emulates `Table` behavior but is not a class. `db.Table` exposes the
       `Table` interface, but is a function which allows omission of metadata.
    """

    def __init__(self, uri='sqlite://',
                 app=None,
                 echo=False,
                 pool_size=None,
                 pool_timeout=None,
                 pool_recycle=None,
                 convert_unicode=True,
                 query_cls=BaseQuery,
                 base_cls=BaseModel):

        self.uri = uri
        self.info = make_url(uri)
        self.options = self._cleanup_options(
            echo=echo,
            pool_size=pool_size,
            pool_timeout=pool_timeout,
            pool_recycle=pool_recycle,
            convert_unicode=convert_unicode,
        )

        self.connector = None
        self._engine_lock = threading.Lock()
        self.session = _create_scoped_session(self, query_cls=query_cls)

        self.Model = declarative_base(cls=base_cls, name='Model')

        self.Model.db = self
        self.Model.query = self.session.query_property()

        if app is not None:
            self.init_app(app)

        _include_sqlalchemy(self)

    def _cleanup_options(self, **kwargs):
        options = dict([
            (key, val)
            for key, val in kwargs.items()
            if val is not None
        ])
        return self._apply_driver_hacks(options)

    def _apply_driver_hacks(self, options):
        if "mysql" in self.info.drivername:
            self.info.query.setdefault('charset', 'utf8')
            options.setdefault('pool_size', 10)
            options.setdefault('pool_recycle', 7200)
        elif self.info.drivername == 'sqlite':
            no_pool = options.get('pool_size') == 0
            memory_based = self.info.database in (None, '', ':memory:')
            if memory_based and no_pool:
                raise ValueError(
                    'SQLite in-memory database with an empty queue'
                    ' (pool_size = 0) is not possible due to data loss.'
                )
        return options

    def init_app(self, app):
        """This callback can be used to initialize an application for the
        use with this database setup. In a web application or a multithreaded
        environment, never use a database without initialize it first,
        or connections will leak.
        """
        if not hasattr(app, 'databases'):
            app.databases = []
        if isinstance(app.databases, list):
            if self in app.databases:
                return
            app.databases.append(self)

        def shutdown(response=None):
            self.session.remove()
            return response

        def rollback(error=None):
            try:
                self.session.rollback()
            except Exception:
                pass

        self.set_flask_hooks(app, shutdown, rollback)

    def set_flask_hooks(self, app, shutdown, rollback):
        if hasattr(app, 'after_request'):
            app.after_request(shutdown)
        if hasattr(app, 'on_exception'):
            app.on_exception(rollback)

    @property
    def engine(self):
        """Gives access to the engine. """
        with self._engine_lock:
            connector = self.connector
            if connector is None:
                connector = EngineConnector(self)
                self.connector = connector
            return connector.get_engine()

    @property
    def metadata(self):
        """Proxy for Model.metadata"""
        return self.Model.metadata

    @property
    def query(self):
        """Proxy for session.query"""
        return self.session.query

    def add(self, *args, **kwargs):
        """Proxy for session.add"""
        return self.session.add(*args, **kwargs)

    def flush(self, *args, **kwargs):
        """Proxy for session.flush"""
        return self.session.flush(*args, **kwargs)

    def commit(self):
        """Proxy for session.commit"""
        return self.session.commit()

    def rollback(self):
        """Proxy for session.rollback"""
        return self.session.rollback()

    def create_all(self):
        """Creates all tables. """
        self.Model.metadata.create_all(bind=self.engine)

    def drop_all(self):
        """Drops all tables. """
        self.Model.metadata.drop_all(bind=self.engine)

    def reflect(self, meta=None):
        """Reflects tables from the database. """
        meta = meta or MetaData()
        meta.reflect(bind=self.engine)
        return meta

    @staticmethod
    def get_dataframe(query):
        """ Converts a query into a Pandas DataFrame """
        return pd.read_sql(query.statement, query.session.bind)

    def __repr__(self):
        return "<SQLAlchemy('{0}')>".format(self.uri)



from flask_wtf import FlaskForm as _FlaskForm
from wtforms_alchemy import model_form_factory as _model_form_factory

_BaseModelForm = _model_form_factory(_FlaskForm)


def create_model_form(db):
    class ModelForm(_BaseModelForm):
        @classmethod
        def get_session(cls):
            return db.session

    return ModelForm



from __future__ import unicode_literals

import io
import sys

from sqlalchemy import inspect
from sqlalchemy.ext.declarative import DeclarativeMeta

from reprlib import Repr as _Repr


__all__ = ['Repr', 'PrettyRepr', 'RepresentableBase',
           'PrettyRepresentableBase']


class Repr(_Repr):
    def repr(self, obj):
        if isinstance(obj.__class__, DeclarativeMeta):
            return self.repr_Base(obj, self.maxlevel)
        if sys.version_info < (3,):
            return _Repr.repr(self, obj)
        else:
            return super(Repr, self).repr(obj)

    def repr_Base(self, obj, level):
        class_repr = self._repr_class(obj, level)
        attrs_repr = self._repr_attrs(obj, level)
        return f'{class_repr}({attrs_repr})'

    def _repr_class(self, obj, level):
        return obj.__class__.__name__

    def _repr_attrs(self, obj, level):
        represented_attrs = []
        for attr in self._iter_attrs(obj):
            represented_attr = self._repr_attr(attr, level)
            represented_attrs.append(represented_attr)
        return ', '.join(represented_attrs)

    def _repr_attr(self, obj, level):
        attr_name, attr_value = obj
        if hasattr(attr_value, 'isoformat'):
            return '%s=%r' % (attr_name, attr_value.isoformat())
        return '%s=%r' % (attr_name, attr_value)

    def _iter_attrs(self, obj):
        attr_names = inspect(obj.__class__).columns.keys()
        for attr_name in attr_names:
            yield (attr_name, getattr(obj, attr_name))


class PrettyRepr(Repr):
    def __init__(self, *args, **kwargs):
        indent = kwargs.pop('indent', None)
        if indent is None:
            self.indent = ' ' * 4
        else:
            self.indent = indent
        if sys.version_info < (3,):
            Repr.__init__(self, *args, **kwargs)
        else:
            super(PrettyRepr, self).__init__(*args, **kwargs)

    def repr_Base(self, obj, level):
        output = io.StringIO()
        output.write('<%s' % self._repr_class(obj, level))
        is_first_attr = True
        for attr in self._iter_attrs(obj):
            if not is_first_attr:
                output.write(',')
            is_first_attr = False
            represented_attr = self._repr_attr(attr, level)
            output.write('\n' + self.indent + represented_attr)
        output.write('>')
        return output.getvalue()


_shared_repr = Repr()
_shared_pretty_repr = PrettyRepr()


class RepresentableBase(object):
    def __repr__(self):
        return _shared_repr.repr(self)


class PrettyRepresentableBase(object):
    def __repr__(self):
        return _shared_pretty_repr.repr(self)



import datetime
import json
from typing import Any, Dict, List

import arrow
import inflection
import pandas as pd
import sqlalchemy_utils as sa_utils
from sqlalchemy import *

from .repr import RepresentableBase


class ModelTableNameDescriptor:
    """
    Create the table name if it doesn't exist.
    """

    def __get__(self, obj, type):
        tablename = type.__dict__.get('__tablename__')
        if not tablename:
            tablename = inflection.underscore(type.__name__)
            setattr(type, '__tablename__', tablename)
        return tablename


class BaseModel(RepresentableBase):
    """
    Baseclass for custom user models.
    """
    __tablename__ = ModelTableNameDescriptor()
    __primary_key__ = "id"  # String

    def __iter__(self):
        """Returns an iterable that supports .next()
        so we can do dict(sa_instance).
        """
        for k in self.__dict__.keys():
            if not k.startswith('_'):
                yield (k, getattr(self, k))

    def to_dict(self) -> Dict:
        """
        Return an entity as dict
        :returns dict:
        """
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}

    def to_json(self) -> str:
        """
        Convert the entity to JSON
        :returns str:
        """
        data = {}
        for k, v in self.to_dict().items():
            if isinstance(v, (datetime.datetime, sa_utils.ArrowType, arrow.Arrow)):
                v = v.isoformat()
            data[k] = v
        return json.dumps(data)

    @classmethod
    def get(cls, pk):
        """
        Select entry by its primary key. It must be define as
        __primary_key__ (string)
        """
        return cls.query.filter(getattr(cls, cls.__primary_key__) == pk).first()

    @classmethod
    def create(cls, **kwargs):
        """
        To create a new record
        :returns object: The new record
        """
        record = cls(**kwargs).save()
        return record

    def update(self, **kwargs):
        """
        Update an entry
        """
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.save()
        return self

    def save(self):
        """
        Shortcut to add and save + rollback
        """
        try:
            self.db.add(self)
            self.db.commit()
            return self
        except Exception as e:
            self.db.rollback()
            raise

    def delete(self):
        """
        Delete a record
        """
        try:
            self.db.session.delete(self)
            return self.db.commit()
        except Exception as e:
            self.db.rollback()
            raise

    def is_valid(self) -> bool:
        """ Takes an sqlalchemy orm object and will return True if it is valid """
        if not self._sa_instance_state.transient:
            raise ValueError("Can not validate existing objects")

        # check to see if obj already exists in database as logic is not able to validate existing entries
        pk = getattr(self, self.__primary_key__)
        if self.get(pk) is not None:
            raise ValueError("Can not validate existing objects")

        save = self.db.session.begin_nested()
        dummy = self.db.session.begin_nested()
        try:
            dummy.session.add(self)
            dummy.commit()   # will raise IntegrityError if not valid

            # tidy up database
            save.rollback()

            # make obj useable again
            setattr(self, self.__primary_key__, None)
            return True
        except Exception as e:
            # make db.session useable again
            save.rollback()
            return False

    @classmethod
    def bulk_insert(cls, mappings: List[Dict], **kwargs):
        """
        Insert a list of dicts to the database.

        Not as fast as `insert_dataframe()` but can be faster than converting list to DataFrame then inserting
        """
        save = cls.db.session.begin_nested()
        try:
            cls.db.session.bulk_insert_mappings(cls, mappings, **kwargs)
            cls.db.commit()
            return True
        except Exception as e:
            save.rollback()
            raise e

    @classmethod
    def insert_dataframe(cls, df: pd.DataFrame):
        """ Insert a Pandas dataframe into the database (fast) """
        save = cls.db.session.begin_nested()
        try:
            df.to_sql(cls.__tablename__, con=cls.db.engine, if_exists='append', index=False)
            return True
        except Exception as e:
            save.rollback()
            raise e



from setuptools import find_packages, setup
import pathlib

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="SqlAlchemy_Tools",
    version="0.1.0-b0",
    include_package_data=True,
    packages=find_packages(),

    install_requires=[
        "sqlalchemy==1.3.20",
        "sqlalchemy-repr==0.0.2",
        "pandas==1.1.4",
        "flask-wtf==0.14.3",
        "wtforms_alchemy==0.17.0",
    ],

    author="Andy Everitt",
    author_email="andreweveritt@e3d-online.com",
    description="Wrapper to make using SqlAlchemy easier, thread safe, helper methods etc.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/AndyEveritt/SqlAlchemyTools",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)
